from datetime import datetime

from freezegun import freeze_time

from schedrem.config import ScheduleConfig
from schedrem.manager import ScheduleManager

DUMMY_NOW = datetime(2024, 10, 28, 0, 0, 0)
DUMMY_WEEKDAYNAMES = [
    ["mon", "tue", "wed", "thu", "fri", "sat", "sun"],
    ["月", "火", "水", "木", "金", "土", "日"],
]


def dummy_week_nums(weekdays: list[str] | None) -> list[int] | None:
    """Convert a list of weekday names to their corresponding weekday numbers.

    According to date.weekday(), Monday is 0 and Sunday is 6.
    """
    if not weekdays:
        return None

    name_to_index: dict[str, int] = {
        name: idx
        for idx, names in enumerate(zip(*DUMMY_WEEKDAYNAMES, strict=False))
        for name in names
    }
    return [name_to_index[w] for w in weekdays if w in name_to_index]


@freeze_time(DUMMY_NOW)
def test_case_1():
    schedule = {"time": {"hour": 1}, "message": "test"}
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2024, 10, 28, 1, 0, 0)


@freeze_time(DUMMY_NOW)
def test_case_2():
    schedule = {
        "time": {"hour": 1},
        "message": "test",
        "wait": {"year": 2025, "month": 2},
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2025, 2, 1, 1, 0, 0)


@freeze_time(DUMMY_NOW)
def test_case_3():
    schedule = {
        "time": {"year": [2023, 3000, 3001], "hour": 1, "minute": [5, 10, 20]},
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(3000, 1, 1, 1, 5, 0)


@freeze_time(DUMMY_NOW)
def test_case_4():
    schedule = {
        "time": {"year": 5000, "month": [2, 3], "hour": 1, "minute": [5, 10, 20]},
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(5000, 2, 1, 1, 5, 0)


@freeze_time(DUMMY_NOW)
def test_case_5():
    schedule = {
        "time": {
            "year": 5000,
            "month": [2, 3],
            "day": 5,
            "hour": 1,
            "minute": [5, 10, 20],
            "weekday": "Fri",
        },
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) is None


@freeze_time(DUMMY_NOW)
def test_case_6():
    schedule = {
        "time": {
            "year": 5000,
            "month": [2, 3],
            "day": 5,
            "hour": 1,
            "minute": [5, 10, 20],
            "weekday": ["weD", "Fri"],
        },
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(5000, 2, 5, 1, 5)


@freeze_time(DUMMY_NOW)
def test_case_7():
    schedule = {
        "time": {
            "day": 13,
            "weekday": "Fri",
        },
        "wait": {
            "year": 5000,
        },
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(5000, 6, 13, 0, 0)


@freeze_time(DUMMY_NOW)
def test_case_8():
    schedule = {
        "time": {
            "day": 13,
            "weekday": "Fri",
        },
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2024, 12, 13, 0, 0)


@freeze_time(DUMMY_NOW)
def test_case_9():
    schedule = {
        "time": {"month": [10, 12], "minute": [1, 59]},
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2024, 10, 28, 0, 1)


@freeze_time(DUMMY_NOW)
def test_case_10():
    schedule = {
        "time": {},
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2024, 10, 28, 0, 1)


@freeze_time(DUMMY_NOW)
def test_case_11():
    schedule = {
        "time": {},
        "message": "test",
        "wait": {"year": 2025, "month": 2},
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2025, 2, 1, 0, 0)


@freeze_time(DUMMY_NOW)
def test_case_12():
    schedule = {
        "time": {"minute": [3], "dow": "wed"},
        "message": "test",
        "wait": {"year": 2025, "month": 2},
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2025, 2, 5, 0, 3)


@freeze_time(DUMMY_NOW)
def test_case_13():
    schedule = {
        "time": {"year": [9998], "dow": "wed"},
        "message": "test",
        "wait": {"year": 9999, "month": 2},
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) is None


@freeze_time(DUMMY_NOW)
def test_case_14():
    schedule = {
        "time": {"year": [2015], "month": 10, "day": [21]},
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) is None


@freeze_time(DUMMY_NOW)
def test_case_15():
    schedule = {
        "time": {"year": [2015, 2024, 2025], "month": 10, "minute": [1, 5]},
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2024, 10, 28, 0, 1)


@freeze_time(DUMMY_NOW)
def test_case_16():
    schedule = {
        "time": {"year": 2025, "month": 10, "dow": ["mon", "金"]},
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2025, 10, 3, 0, 0)


@freeze_time(DUMMY_NOW)
def test_case_17():
    schedule = {
        "time": {
            "year": [],
            "month": [],
            "dow": ["mon", "金"],
            "hour": [],
            "minute": [3],
        },
        "message": "test",
    }
    sch = ScheduleConfig(**schedule)
    schman = ScheduleManager(sch, dummy_week_nums, None)
    assert schman.nearest_future(sch.time) == datetime(2024, 10, 28, 0, 3)



from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    NonNegativeFloat,
    ValidationInfo,
    field_validator,
    model_validator,
)


class WaitConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    year: int
    month: int = 1
    day: int = 1
    hour: int = 0
    minute: int = 0

    @field_validator("month", "day", "hour", "minute")
    @classmethod
    def validate_time(cls, v, info: ValidationInfo):
        field_name = str(info.field_name)
        min_val = {"month": 1, "day": 1, "hour": 0, "minute": 0}[field_name]
        max_val = {"month": 12, "day": 31, "hour": 23, "minute": 59}[field_name]
        if v < min_val or v > max_val:
            msg = f"{field_name} must be between {min_val} and {max_val}."
            raise ValueError(msg)
        return v


class TimeConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    year: list[int] | None = None
    month: list[int] | None = None
    day: list[int] | None = None
    weekday: list[str] | None = Field(
        default=None,
        validation_alias=AliasChoices("weekday", "dow"),
    )
    hour: list[int] | None = None
    minute: list[int] | None = None

    @field_validator("*", mode="before")
    @classmethod
    def ensure_list(cls, v):
        if type(v) in (int, str):
            return [v]
        return v

    @field_validator("month", "day", "hour", "minute")
    @classmethod
    def validate_time(cls, v, info: ValidationInfo):
        field_name = str(info.field_name)
        min_val = {"month": 1, "day": 1, "hour": 0, "minute": 0}[field_name]
        max_val = {"month": 12, "day": 31, "hour": 23, "minute": 59}[field_name]
        if type(v) is list and any(w < min_val or w > max_val for w in v):
            msg = f"{field_name} must be between {min_val} and {max_val}."
            raise ValueError(msg)
        return v

    @field_validator("weekday")
    @classmethod
    def lower_weekday(cls, v):
        """Weekday name matching is case insensitive."""
        if type(v) is list:
            return [w.lower() for w in v]
        return v

    @field_validator("weekday")
    @classmethod
    def weekday_in_weekdaynames(cls, v, info: ValidationInfo):
        """Check weekday is in weekdaynames."""
        if v is None or not info.context:
            return v
        wdn = info.context.get("weekdaynames")
        idx = info.context.get("schedule_idx")
        for w in v:
            if all(w not in aweek for aweek in wdn):
                msg = (
                    f'weekday name "{w}" in schedules.{idx}'
                    f" is not in weekdaynames {wdn}."
                )
                raise ValueError(msg)
        return v


class ActionConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    yesno: str | None = None
    command: str | None = Field(
        default=None,
        validation_alias=AliasChoices("command", "cmd"),
    )
    message: str | None = Field(
        default=None,
        validation_alias=AliasChoices("message", "msg"),
    )
    sound: str | bool | None = None
    font: str | None = None

    @model_validator(mode="after")
    def validate_actions(self):
        """Check if at least one action is defined."""
        if self.yesno is None and self.command is None and self.message is None:
            msg = "No action specified."
            raise ValueError(msg)
        return self

    @field_validator("sound")
    @classmethod
    def validate_sound(cls, v):
        """Accept WAV only."""
        if type(v) is str and not v.lower().endswith(".wav"):
            msg = "The only accepted format for audio files is WAV."
            raise ValueError(msg)
        return v


class ScheduleConfig(ActionConfig):
    model_config = ConfigDict(extra="forbid")

    description: str = Field(
        default="",
        validation_alias=AliasChoices("description", "desc"),
    )
    time: TimeConfig = TimeConfig()
    wait: WaitConfig | None = None
    # additional delay (in seconds) to be applied to each scheduled time
    delay: NonNegativeFloat = 0.0
    enabled: bool = True

    @field_validator("time")
    @classmethod
    def validate_weekday(cls, v, info: ValidationInfo):
        """Let TimeConfig check weekday is in weekdaynames."""
        if not info.context:
            return v
        wdn = info.context.get("weekdaynames")
        idx = info.context.get("schedule_idx")
        TimeConfig.model_validate(
            dict(v),
            context={"weekdaynames": wdn, "schedule_idx": idx},
        )
        return v


class SchedremConfig(BaseModel):
    disabled: bool = False
    weekdaynames: list[list[str]] = [["mon", "tue", "wed", "thu", "fri", "sat", "sun"]]
    schedules: list[ScheduleConfig] = []
    font: str | None = None

    @field_validator("weekdaynames")
    @classmethod
    def lower_weekdaynames(cls, v):
        """Weekday name matching is case insensitive."""
        return [[x.lower() for x in w] for w in v]

    @field_validator("weekdaynames")
    @classmethod
    def validate_weekdaynames(cls, v):
        """Check that weekdaynames has lists of unique 7 elements."""
        number_of_dow = 7
        for weekdays in v:
            if len(weekdays) > len(set(weekdays)):
                msg = "weekdaynames should have lists of unique elements."
                raise ValueError(msg)
            if len(weekdays) != number_of_dow:
                msg = "weekdaynames should have lists of 7 elements."
                raise ValueError(msg)
        return v

    @model_validator(mode="after")
    def validate_weekday(self):
        """Let TimeConfig in ScheduleConfig check weekday is in weekdaynames."""
        for i, schedule in enumerate(self.schedules):
            ScheduleConfig.model_validate(
                dict(schedule),
                context={"weekdaynames": self.weekdaynames, "schedule_idx": i},
            )

        return self



import asyncio
import logging
import subprocess
import sys
from collections.abc import Callable
from datetime import datetime, timedelta
from pathlib import Path

import psutil
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer
from yaml import safe_load

from .config import SchedremConfig, ScheduleConfig, TimeConfig


class SchedremManager:
    def __init__(self, yaml_path: Path) -> None:
        self.coros: list = []
        self.tasks: list = []
        self.observer = Observer()
        self.observer.schedule(
            SchedremEventHandler(yaml_path, self),
            str(yaml_path.resolve().parent),
            recursive=False,
        )
        self.config = SchedremConfig(**self.load_yaml(yaml_path))
        if self.config.disabled:
            logging.debug("Schedrem is disabled.\n")
            # terminate all existing action processes when disabled
            for proc in psutil.process_iter(["name", "cmdline"]):
                try:
                    name = proc.info["name"]
                    cmdline = proc.info["cmdline"]
                    if (
                        name in ("schedrem", "schedrem.exe")
                        and cmdline is not None
                        and "--action" in cmdline
                    ):
                        proc.terminate()
                except Exception as e:
                    logging.debug("Failed to terminate process: %s", e)

        else:
            self.set_schedules()

    def load_yaml(self, yaml_path: Path) -> dict:
        with yaml_path.open(encoding="utf-8", errors="replace") as file:
            return safe_load(file) or {}

    def week_nums(self, weekdays: list[str] | None) -> list[int] | None:
        """Convert a list of weekday names to their corresponding weekday numbers.

        According to date.weekday(), Monday is 0 and Sunday is 6.
        """
        if not weekdays:
            return None

        name_to_index: dict[str, int] = {
            name: idx
            for idx, names in enumerate(zip(*self.config.weekdaynames, strict=False))
            for name in names
        }
        return [name_to_index[w] for w in weekdays if w in name_to_index]

    def set_schedules(self) -> None:
        self.sched_mans = [
            ScheduleManager(sch, self.week_nums, self.config.font)
            for sch in self.config.schedules
        ]
        self.coros = [sched.standby for sched in self.sched_mans]

    async def cancel_awaiter(self) -> None:
        """Just a sleep loop waiting for cancellation."""
        while True:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break

    def run(self) -> None:
        logging.debug("weekdaynames: %s\n", self.config.weekdaynames)

        async def create_tasks() -> None:
            async with asyncio.TaskGroup() as tg:
                self.tasks = [tg.create_task(coro()) for coro in self.coros]
                self.tasks.append(tg.create_task(self.cancel_awaiter()))

        self.observer.start()
        try:
            asyncio.run(create_tasks())
        except KeyboardInterrupt:
            sys.exit(130)
        finally:
            logging.debug("Tasks have been cancelled!\n")
            self.observer.stop()
            self.observer.join()


class ScheduleManager:
    def __init__(
        self,
        sconf: ScheduleConfig,
        week_nums: Callable,
        font: str | None,
    ) -> None:
        self.sconf = sconf
        self.sconf.font = self.sconf.font or font
        self.week_nums = week_nums

    async def standby(self) -> None:
        while True:
            """No time key should be set in a schedule
            if it should be triggered every minute.
            It means the same as * * * * * in crontab.
            """
            if not self.sconf.enabled:
                return
            nexttime = self.nearest_future(self.sconf.time)
            if not nexttime:
                return
            nexttime += timedelta(seconds=self.sconf.delay)
            waittime = (nexttime - datetime.now()).total_seconds()
            logging.debug(self.sconf)
            logging.debug(nexttime)
            logging.debug("%s seconds left\n", waittime)

            try:
                await asyncio.sleep(waittime)
                while datetime.now() < nexttime:
                    """Adjust the gap between monotonic and real-time clocks
                    if waittime is not enough for nexttime.
                    """
                    await asyncio.sleep(0.1)
                action_arg = self.sconf.model_dump_json(
                    include={
                        "yesno",
                        "command",
                        "message",
                        "sound",
                        "font",
                    },
                )
                subprocess.Popen(["schedrem", "--action", action_arg])
            except asyncio.CancelledError:
                break

    def nearest_future(self, tconf: TimeConfig) -> datetime | None:
        now = datetime.now().replace(second=0, microsecond=0)
        now += timedelta(minutes=1)  # trigger from next minute

        if self.sconf.wait:
            candidate = datetime(**self.sconf.wait.model_dump())
            candidate = max(candidate, now)
        else:
            candidate = now

        if tconf.year:
            if candidate.year > max(tconf.year):
                return None
            if candidate.year not in tconf.year:
                candidate = datetime(
                    min([y for y in tconf.year if y > candidate.year]),
                    min(tconf.month) if tconf.month else 1,
                    min(tconf.day) if tconf.day else 1,
                )

        if (
            tconf.month
            and candidate.month not in tconf.month
            or tconf.day
            and candidate.day not in tconf.day
        ):
            delta = {"days": 1}
            candidate = datetime(
                candidate.year,
                candidate.month,
                candidate.day,
                min(tconf.hour) if tconf.hour else 0,
                min(tconf.minute) if tconf.minute else 0,
            )
        else:
            delta = {"minutes": 1}

        def time_match(candi: int, setting: list[int] | None) -> bool:
            return not setting or candi in setting

        while not tconf.year or min(tconf.year) <= candidate.year <= max(tconf.year):
            if (
                time_match(candidate.year, tconf.year)
                and time_match(candidate.month, tconf.month)
                and time_match(candidate.day, tconf.day)
                and time_match(candidate.hour, tconf.hour)
                and time_match(candidate.minute, tconf.minute)
                and time_match(candidate.weekday(), self.week_nums(tconf.weekday))
            ):
                return candidate
            candidate += timedelta(**delta)

        return None


class SchedremEventHandler(FileSystemEventHandler):
    def __init__(self, yaml_path: Path, app_manager: SchedremManager) -> None:
        self.yaml_path = yaml_path
        self.app_manager = app_manager

    def is_yaml_path(self, path: str) -> bool:
        return Path(path) == self.yaml_path

    def on_any_event(self, event: FileSystemEvent) -> None:
        if event.event_type in [
            "modified",
            "moved",
            "deleted",
        ] and self.is_yaml_path(str(event.src_path)):
            for task in self.app_manager.tasks:
                task.cancel()



from PySide6.QtCore import QEvent, QObject, Qt, QTimer
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication, QMessageBox


class KeyBlocker(QObject):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.parent = parent

    def eventFilter(self, _, event) -> bool:
        if not self.parent.locked:
            return False

        evtype = event.type()
        if evtype == QEvent.KeyPress and event.key() in (
            Qt.Key_Return,
            Qt.Key_Enter,
            Qt.Key_Escape,
            Qt.Key_Space,
        ):
            return True
        return evtype == QEvent.MouseButtonPress


class BlockingMessageBox(QMessageBox):
    def __init__(
        self,
        icon,
        title,
        message,
        font,
        window_icon,
        buttons=QMessageBox.Ok,
    ) -> None:
        super().__init__()

        self.locked = True

        self.setWindowFlag(Qt.WindowStaysOnTopHint)
        self.setIcon(icon)
        self.setWindowIcon(QIcon(str(window_icon)))
        self.setWindowTitle(title)
        self.setText(message)
        self.setStandardButtons(buttons)

        font_family, font_size = self.parse_font(font)

        style = (
            f'QMessageBox {{ font-family: "{font_family}"; font-size: {font_size}pt; }}'
        )
        self.setStyleSheet(style)

        blocker = KeyBlocker(self)
        for btn in self.buttons():
            btn.installEventFilter(blocker)

        QTimer.singleShot(500, lambda: setattr(self, "locked", False))

        self.raise_()
        self.activateWindow()

    def parse_font(self, font: str | None) -> tuple[str, str]:
        default = ("Arial", "15")
        if type(font) is not str:
            return default
        parts = font.rsplit(maxsplit=1)
        if len(parts) == 2:  # noqa: PLR2004
            return tuple(parts)  # type: ignore[return-value]
        return default


def _app() -> QApplication:
    QApplication.instance() or QApplication()


def _show(
    icon, title, message, font, window_icon, buttons=QMessageBox.Ok
) -> BlockingMessageBox:
    _app()
    return BlockingMessageBox(icon, title, message, font, window_icon, buttons).exec()


def showinfo(title, message, font, window_icon):
    _show(
        QMessageBox.Information,
        title,
        message,
        font,
        window_icon,
        buttons=QMessageBox.Ok,
    )


def showwarning(title, message, font, window_icon):
    _show(
        QMessageBox.Warning, title, message, font, window_icon, buttons=QMessageBox.Ok
    )


def showerror(title, message, font, window_icon):
    _show(
        QMessageBox.Critical, title, message, font, window_icon, buttons=QMessageBox.Ok
    )


def askyesno(title, message, font, window_icon):
    return (
        _show(
            QMessageBox.Question,
            title,
            message,
            font,
            window_icon,
            buttons=QMessageBox.Yes | QMessageBox.No,
        )
        == QMessageBox.Yes
    )



import logging
import os
import subprocess
import sys
import threading
import wave
from argparse import ArgumentParser, Namespace
from pathlib import Path

import pyaudio

from .config import ActionConfig
from .messagebox import askyesno, showerror, showinfo, showwarning


def program_dir() -> Path:
    if getattr(sys, "frozen", False):
        return Path(sys.executable).resolve().parent
    return Path(__file__).resolve().parent


def set_logger(debug: bool) -> None:
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.WARNING,
        format="%(message)s",
    )
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    logging.getLogger("watchdog").setLevel(logging.WARNING)


def take_action(action: ActionConfig) -> int:
    if action.command and not action.yesno:
        proc = subprocess.Popen(action.command, shell=True)
    if action.yesno:
        logging.debug("yesno: %s\n", action.yesno)
        m = Messenger(action.sound, action.font)
        yes = m.yesno(action.yesno)
        if action.command and yes:
            proc = subprocess.Popen(action.command, shell=True)
        if not yes:
            # action.message won't be shown if action.yesno gets "No" response
            return 1
    if action.message:
        logging.debug("message: %s\n", action.message)
        m = Messenger(action.sound, action.font)
        m.message(action.message)
    if action.command:
        proc.wait()
        return proc.returncode
    return 0


class Messenger:
    prg_dir: Path = program_dir()
    default_sound: Path = prg_dir / "assets/alert.wav"
    icon: Path = prg_dir / "assets/schedrem.ico"

    def __init__(
        self,
        sound: str | bool | None = None,
        font: str | None = None,
    ) -> None:
        self.sound_path: Path | None
        if sound is True:
            self.sound_path = self.default_sound
        elif type(sound) is str:
            self.sound_path = Path(os.path.expandvars(sound)).expanduser()
            if not self.sound_path.exists():
                self.sound_path = self.default_sound
        else:
            self.sound_path = None

        self.font = font
        self.keep: bool = True

    def sing(self) -> None:
        if self.sound_path is None:
            return
        self.keep = True
        with wave.open(str(self.sound_path), "rb") as wf:
            sampwidth = wf.getsampwidth()
            framerate = wf.getframerate()
            channels = wf.getnchannels()
            p = pyaudio.PyAudio()
            stream = p.open(
                format=p.get_format_from_width(sampwidth),
                channels=channels,
                rate=framerate,
                output=True,
            )

            chunk = 1024
            while self.keep:
                wf.rewind()
                buffer = wf.readframes(chunk)
                while self.keep and buffer:
                    stream.write(buffer)
                    buffer = wf.readframes(chunk)

            stream.stop_stream()
            stream.close()
        p.terminate()

    def start_singing(self) -> None:
        self.keep = True
        self.song_thread = threading.Thread(target=self.sing, daemon=True)
        self.song_thread.start()

    def stop_singing(self) -> None:
        self.keep = False
        self.song_thread.join()

    def message(self, text: str) -> None:
        self.start_singing()
        showinfo("schedrem", text, self.font, self.icon)
        self.stop_singing()

    def yesno(self, text: str) -> bool:
        ans = False
        self.start_singing()
        ans = askyesno("schedrem", text, self.font, self.icon)
        self.stop_singing()
        return ans

    def warning(self, text: str) -> None:
        showwarning("schedrem", text, self.font, self.icon)

    def error(self, text: str) -> None:
        showerror("schedrem", text, self.font, self.icon)


def error_message(errors: list) -> str:
    messages = []
    for error in errors:
        loc = ".".join([str(i) for i in error.get("loc", [])])
        msg = error.get("msg", "")
        ipt = error.get("input")
        message = f"{loc}\n{msg}\ninput: {ipt}"
        messages.append(message)
    return "\n".join(messages)


def get_args() -> Namespace:
    parser = ArgumentParser(description="Task manager and reminder")
    parser.add_argument(
        "--config",
        type=str,
        help="Path to config file for schedules",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Debug mode",
    )
    parser.add_argument(
        "--action",
        type=str,
        help="JSON string to be executed as an action",
    )
    return parser.parse_args()


def get_config_file() -> Path:
    r"""Search and return path to the config file.

    The files are searched in the order specified below and
    only the first one that is found is read.
    Defaults for Linux/macOS:
    "${XDG_CONFIG_HOME}/schedrem/config.yml"
    "${XDG_CONFIG_HOME}/schedrem/config.yaml"
    "${HOME}/.config/schedrem/config.yml"
    "${HOME}/.config/schedrem/config.yaml"
    Defaults for Windows:
    - "%USERPROFILE%\.config\schedrem\config.yml"
    - "%USERPROFILE%\.config\schedrem\config.yaml"
    - "%APPDATA%\schedrem\config.yml"
    - "%APPDATA%\schedrem\config.yaml"
    """
    if os.name == "posix":
        paths = (
            "${XDG_CONFIG_HOME}/schedrem/config.yml",
            "${XDG_CONFIG_HOME}/schedrem/config.yaml",
            "${HOME}/.config/schedrem/config.yml",
            "${HOME}/.config/schedrem/config.yaml",
        )
    elif os.name == "nt":
        paths = (
            "${USERPROFILE}/.config/schedrem/config.yml",
            "${USERPROFILE}/.config/schedrem/config.yaml",
            "${APPDATA}/schedrem/config.yml",
            "${APPDATA}/schedrem/config.yaml",
        )
    else:
        msg = "OS must be either POSIX or NT."
        raise OSError(msg)

    for p in paths:
        path = Path(os.path.expandvars(p)).resolve()
        if path.exists():
            return path

    msg = f"Config file not found in {paths}."
    raise FileNotFoundError(msg)
